import dearpygui.dearpygui as dpg
from dearpygui_router import navigate

def add_nav_button(route_name: str, label=None, **kwargs):
    label = label or route_name.split("/")[-1]
    dpg.add_button(label=label, callback=lambda: navigate(route_name), **kwargs)


import sys
import types
import pytest
from unittest.mock import MagicMock
from itertools import count

_id_generator = count(1)
def generate_uuid():
    return next(_id_generator)

@pytest.fixture(autouse=True)
def mock_dpg(monkeypatch):
    assert "dearpygui.dearpygui" not in sys.modules, "DPG was imported too early!"
    fake_dpg = MagicMock()
    fake_dpg.generate_uuid = MagicMock(side_effect=generate_uuid)

    # Insert fake module before import happens
    monkeypatch.setitem(sys.modules, "dearpygui", types.SimpleNamespace(dearpygui=fake_dpg))
    monkeypatch.setitem(sys.modules, "dearpygui.dearpygui", fake_dpg)

    yield fake_dpg



import dearpygui.dearpygui as dpg
from dearpygui_router import Router

def screen1():
    dpg.add_text("Screen 1")
    dpg.add_button(label="Go To 2", callback=lambda: router_top.navigate("screen2"))

def screen2():
    dpg.add_text("Screen 2")
    dpg.add_button(label="Go To 1", callback=lambda: router_top.navigate("screen1"))

    with dpg.group(horizontal=True):
        with dpg.child_window(width=200):
            dpg.add_text("Left")
            router_left.add_to_context()
        with dpg.child_window(width=200):
            dpg.add_text("Right")
            router_right.add_to_context()

def left1():
    dpg.add_text("Left 1")
def left2():
    dpg.add_text("Left 2")
def right1():
    dpg.add_text("Right 1")
def right2():
    dpg.add_text("Right 2")

dpg.create_context()
dpg.create_viewport()
dpg.setup_dearpygui()

router_top = Router("top")
router_top.register_screen_function("screen1", screen1, is_initial=True)
router_top.register_screen_function("screen2", screen2)

router_left = Router("left")
router_left.register_screen_function("p1", left1, is_initial=True)
router_left.register_screen_function("p2", left2)

router_right = Router("right")
router_right.register_screen_function("p1", right1, is_initial=True)
router_right.register_screen_function("p2", right2)

router_top.add_nested(router_right)


with dpg.window(tag="_main_window", label="Router Demo"):
    with dpg.group(horizontal=True):
        dpg.add_button(label="Go To 1", callback=lambda: router_top.navigate("screen1"))
        dpg.add_button(label="Go To 2", callback=lambda: router_top.navigate("screen2"))
        dpg.add_button(label="Jump to screen2 right p2", callback=lambda: router_top.navigate("screen2/p2"))
        dpg.add_button(label="Jump to screen2 right p1", callback=lambda: router_top.navigate("screen2/p1"))
    
    router_top.add_to_context()

dpg.show_viewport()
dpg.set_primary_window("_main_window", True)
dpg.start_dearpygui()
dpg.destroy_context()



from typing import Callable
from unittest.mock import MagicMock

def test_router_calls_screen_function_on_nav(mock_dpg):
    from dearpygui_router import Router
    
    router = Router("main")
    screen_fn_1 = MagicMock(spec=Callable)
    screen_fn_2 = MagicMock(spec=Callable)
    router.register_screen_function("screen1", screen_fn_1)
    router.register_screen_function("screen2", screen_fn_2)

    router.navigate("screen1")

    screen_fn_1.assert_called_once()
    screen_fn_2.assert_not_called()

def test_router_navigates_to_initial(mock_dpg):
    from dearpygui_router import Router
    
    router = Router("main")
    screen_fn_1 = MagicMock(spec=Callable)
    screen_fn_2 = MagicMock(spec=Callable)
    router.register_screen_function("screen1", screen_fn_1, is_initial=True)
    router.register_screen_function("screen2", screen_fn_2)

    router.add_to_context()

    screen_fn_1.assert_called_once()
    screen_fn_2.assert_not_called()

def test_router_navigates_to_nested(mock_dpg):
    from dearpygui_router import Router
    
    router1 = Router("main")
    router2 = Router("nested")
    screen_fn_1 = MagicMock(spec=Callable)
    router1.register_screen_function("screen1", screen_fn_1, is_initial=True)

    screen_fn_2 = MagicMock(spec=Callable)
    router2.register_screen_function("screen2", screen_fn_2)
    router1.add_nested(router2)

    router1.navigate("screen1")
    screen_fn_1.assert_called_once()
    screen_fn_2.assert_not_called()

    router1.navigate("screen1/screen2")
    screen_fn_1.assert_called()
    screen_fn_2.assert_called_once()



from typing import List

from dearpygui_router.types import RouterProtocol

_ROUTER_STACK: List[RouterProtocol] = []

def push_router(router: RouterProtocol):
    _ROUTER_STACK.append(router)

def pop_router() -> RouterProtocol:
    return _ROUTER_STACK.pop()

def top_router() -> RouterProtocol:
    if not _ROUTER_STACK:
        return None
    return _ROUTER_STACK[-1]


import dearpygui.dearpygui as dpg
import dearpygui_router as router

class FormStep1:
    def __init__(self):
        self.data = {}

    def add_to_context(self):
        print("Render page 1")
        dpg.add_input_text(label="name", tag="s1_name")
        dpg.add_input_text(label="email", tag="s1_email")
        dpg.add_spacer(height=10)
        dpg.add_button(label="Next", width=100, callback=self._on_btn_next)

    def _on_btn_next(self):
        self.data["name"] = dpg.get_value("s1_name")
        self.data["email"] = dpg.get_value("s1_email")
        router.navigate("step2")

class FormStep2:
    def __init__(self):
        self.data = {}

    def add_to_context(self):
        dpg.add_combo(items=["Cat", "Dog", "Gerbil"], label="Pet Type", tag="s2_pet")
        dpg.add_input_int(label="Pet Age", tag="s2_age")
        dpg.add_spacer(height=10)
        dpg.add_button(label="Next", width=100, callback=self._on_btn_next)

    def _on_btn_next(self):
        self.data["petType"] = dpg.get_value("s2_pet")
        self.data["petAge"] = dpg.get_value("s2_age")
        router.navigate("step3")

def summary_page():
    with dpg.group():
        dpg.add_text("Summary of your responses")
        dpg.add_spacer(height=10)

        with router.router():
            with router.screen("you", is_initial=True):
                name = step1.data["name"]
                email = step1.data["email"]
                dpg.add_text("You", color=(30, 220, 30))
                dpg.add_text(f"Your name is {name} and your email is {email}")
                dpg.add_spacer(height=10)
                dpg.add_button(label="Your Pet", width=100, callback=lambda: router.navigate("step3/pet"))

            with router.screen("pet"):
                pet_type = step2.data["petType"]
                age = step2.data["petAge"]
                dpg.add_text("Your Pet", color=(30, 220, 30))
                dpg.add_text(f"You have a pet {pet_type} that is {age} years old")
                dpg.add_spacer(height=10)
                dpg.add_button(label="You", width=100, callback=lambda: router.navigate("step3/you"))

step1 = FormStep1()
step2 = FormStep2()

dpg.create_context()
dpg.create_viewport()
dpg.setup_dearpygui()

with dpg.window(tag="_main_window", label="Router Demo: Form"):
    dpg.add_text("Please fill out this form")
    dpg.add_spacer(height=30)

    with router.router():
        router.add_screen_object("step1", step1, is_initial=True)
        router.add_screen_object("step2", step2)
        router.add_screen_function("step3", summary_page)
    

dpg.show_viewport()
dpg.set_primary_window("_main_window", True)
dpg.start_dearpygui()
dpg.destroy_context()


import dearpygui.dearpygui as dpg
from dearpygui_router import Router

def screen1():
    dpg.add_text("Screen 1")
    dpg.add_button(label="Go To 2", callback=lambda: router.navigate("screen2"))

def screen2():
    dpg.add_text("Screen 2")
    dpg.add_button(label="Go To 1", callback=lambda: router.navigate("screen1"))

dpg.create_context()
dpg.create_viewport()
dpg.setup_dearpygui()

router = Router("top")
router.register_screen_function("screen1", screen1, is_initial=True)
router.register_screen_function("screen2", screen2)

with dpg.window(tag="_main_window", label="Router Demo"):
    with dpg.group(horizontal=True):
        dpg.add_button(label="Go To 1", callback=lambda: router.navigate("screen1"))
        dpg.add_button(label="Go To 2", callback=lambda: router.navigate("screen2"))
    
    router.add_to_context()

dpg.show_viewport()
dpg.set_primary_window("_main_window", True)
dpg.start_dearpygui()
dpg.destroy_context()



import dearpygui.dearpygui as dpg
from typing import Callable, Optional

from dearpygui_router.screen import Screen
from dearpygui_router.stack import pop_router, push_router, top_router
from dearpygui_router.router import Router

_ROOT_ROUTERS = []

def root_router() -> Router:
    if not _ROOT_ROUTERS:
        raise RuntimeError("No router defined")
    return _ROOT_ROUTERS[0]

def navigate(path: str):
    router = root_router()
    router.navigate(path)

def router(tag: Optional[str] = None):
    """Start a router context"""
    router_id = tag or dpg.generate_uuid()
    r = Router(router_id)
    return _RouterContext(r)

def screen(route_name: str, is_initial=False):
    """Start a screen context"""
    router = top_router()
    if not router:
        raise RuntimeError("screen() must be called inside router context")
    
    return _ScreenContext(router, route_name, show=is_initial)

def add_screen_object(route_name, obj: Screen, is_initial=False):
    """Add a screen object to the current router context"""
    router = top_router()
    if not router:
        raise RuntimeError("add_screen() must be called inside router context")
    
    router.register_screen_object(route_name, obj, is_initial)

def add_screen_function(route_name, fn: Callable, is_initial=False):
    """Add a screen object to the current router context"""
    router = top_router()
    if not router:
        raise RuntimeError("add_screen() must be called inside router context")
    
    router.register_screen_function(route_name, fn, is_initial)

class _RouterContext:
    def __init__(self, router: Router):
        self.router = router

    def __enter__(self):
        parent = top_router()
        if parent:
            parent.add_nested(self.router)
        else:
            _ROOT_ROUTERS.append(self.router)

        push_router(self.router)
        self.router.add_to_context()
        dpg.push_container_stack(self.router.container_tag)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pop_router()
        dpg.pop_container_stack()
        self.router.home()

class _ScreenContext:
    def __init__(self, router: Router, route_name: str, show=False):
        self.router = router
        self.route_name = route_name
        self.group_id = dpg.generate_uuid()
        self.show = show

    def __enter__(self):
        group = dpg.add_group(tag=self.group_id, show=self.show)
        dpg.push_container_stack(group)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.router.register_screen_item(self.route_name, self.group_id, is_initial=self.show)
        dpg.pop_container_stack()



from dearpygui_router.context import (
    navigate,
    router,
    screen,
    add_screen_function,
    add_screen_object
)
from dearpygui_router.router import Router
from dearpygui_router.screen import Screen
from dearpygui_router.ui import add_nav_button

__all__ = [
    navigate,
    router,
    screen,
    add_screen_function,
    add_screen_object,
    Router,
    Screen,
    add_nav_button
]



from typing import Callable, Protocol, Union

from dearpygui_router.screen import Screen, ScreenHandler

Tag = Union[int, str]

class RouterProtocol(Protocol):
    def register(self, route_name: str, screen: ScreenHandler, is_initial=False): ...
    def register_screen_function(self, route_name: str, fn: Callable, is_initial=False): ...
    def register_screen_object(self, route_name: str, obj: Screen, is_initial=False): ...
    def register_screen_item(self, route_name: str, item_tag: Tag, is_initial=False): ...
    def register_screen_factory(self, route_name: str, factory: Callable, is_initial=False): ...
    def navigate(self, route_name: str) -> None: ...
    def home(self): ...
    def add_to_context(self) -> None: ...
    def add_nested(self, router: "RouterProtocol") -> None: ...


from typing import Dict, Optional, Union, Callable
import dearpygui.dearpygui as dpg

from dearpygui_router.stack import pop_router, push_router
from dearpygui_router.screen import (
    Screen, 
    ScreenFactory, 
    ScreenFunction, 
    ScreenHandler, 
    ScreenInstance, 
    ScreenItemTag
)
from dearpygui_router.types import RouterProtocol

SPECIAL_CHARACTERS = ['/', ':', '#', '*']

Tag = Union[int, str]

class Router(RouterProtocol):
    def __init__(self, name:str):
        self.name = name
        self.container_tag = dpg.generate_uuid()
        self.current_screen:Optional[ScreenHandler] = None
        self.current_route:Optional[str] = None
        self.initial_screen:Optional[str] = None

        self.route_handlers: Dict[str, ScreenHandler] = {}
        self.nested_router: Optional[Router] = None

    def register(self, route_name: str, screen: ScreenHandler, is_initial=False):
        """Register a screen at a named route

        Args:
            route_name (str): Route name. Cannot contain special characters: ['/', ':', '#', '*']
            screen (ScreenHandler): Many types of ScreenHandlers are available based on how the UI is constructed
            is_initial: if True, this screen will be shown on initial render 
        """
        print(f"Register route: {route_name}")
        if route_name in self.route_handlers:
            raise ValueError(f"Route [{route_name}] already exists on this router")
        
        for char in SPECIAL_CHARACTERS:
            if char in route_name:
                raise ValueError(f"Route name cannot contain '{char}'. Character is reserved")
        
        self.route_handlers[route_name] = screen
        
        if is_initial:
            print(f"Set initial route: {route_name}")
            self.initial_screen = route_name

    def register_screen_function(self, route_name: str, fn: Callable, is_initial=False):
        """Register a screen at a named route

        Args:
            route_name (str): Route name. Cannot contain special characters: ['/', ':', '#', '*']
            screen (Callable): A function the sets up the DPG UI elements
            is_initial: if True, this screen will be shown on initial render 
        """
        self.register(route_name, ScreenFunction(fn), is_initial)

    def register_screen_object(self, route_name: str, obj: Screen, is_initial=False):
        """Register a screen at a named route

        Args:
            route_name (str): Route name. Cannot contain special characters: ['/', ':', '#', '*']
            screen (Screen): An object that implements the Screen protocol
            is_initial: if True, this screen will be shown on initial render 
        """
        self.register(route_name, ScreenInstance(obj), is_initial)

    def register_screen_item(self, route_name: str, item_tag: Tag, is_initial=False):
        """Register a screen at a named route

        Args:
            route_name (str): Route name. Cannot contain special characters: ['/', ':', '#', '*']
            screen (int | str): A DPG item, like a group or child_window
            is_initial: if True, this screen will be shown on initial render 
        """
        screen = ScreenItemTag(item_tag)
        self.register(route_name, screen, is_initial)
        if is_initial:
            self.current_screen = screen
            self.current_route = route_name

    def register_screen_factory(self, route_name: str, factory: Callable, is_initial=False):
        """Register a screen at a named route

        Args:
            route_name (str): Route name. Cannot contain special characters: ['/', ':', '#', '*']
            screen (Callable): A function that creates and returns a Screen instance
            is_initial: if True, this screen will be shown on initial render 
        """
        self.register(route_name, ScreenFactory(factory), is_initial)

    def navigate(self, route_name: str) -> None:
        route_first, _, route_rest = route_name.partition("/")
        route_rest = route_rest or None

        print("nav", route_first, route_rest)

        if route_first not in self.route_handlers:
            raise ValueError(f"Route '{route_first}' not found on this router.")

        screen = self.route_handlers[route_first]

        # Hide or remove previous screen
        if self.current_screen:
            # TODO: add a cleanup/remove hook to `Screen`
            pass

        if self.current_screen is not None:
            self.current_screen.hide(self.container_tag)


        push_router(self)
        screen.show(self.container_tag)
        self.current_screen = screen
        self.current_route = route_first

        if route_rest and self.nested_router:
            self.nested_router.navigate(route_rest)
        pop_router()

    def home(self):
        if self.initial_screen is not None:
            self.navigate(self.initial_screen)

    def add_to_context(self) -> None:
        print("Add router UI to context")
        with dpg.child_window(tag=self.container_tag, border=False):
            pass

        if self.initial_screen is not None:
            self.navigate(self.initial_screen)

    def add_nested(self, router: "Router") -> None:
        self.nested_router = router


import abc
from typing import Callable, Protocol, Union, runtime_checkable
import dearpygui.dearpygui as dpg

Tag = Union[int, str]

@runtime_checkable
class Screen(Protocol):
    def add_to_context(self) -> None: ...

class ScreenHandler(abc.ABC):
    def show(self, container: Tag) -> None: ...
    def hide(self, container: Tag) -> None: ...

class ScreenFunction(ScreenHandler):
    def __init__(self, fn: Callable):
        self.fn = fn

    def show(self, container: Tag) -> None:
        with dpg.group(parent=container):
            self.fn()

    def hide(self, container: Tag) -> None:
        dpg.delete_item(container, children_only=True)

class ScreenInstance(ScreenHandler):
    def __init__(self, screen: Screen):
        self.instance = screen

    def show(self, container: Tag) -> None:
        print("Show screen instance")
        with dpg.group(parent=container):
            self.instance.add_to_context()

    def hide(self, container: Tag) -> None:
        dpg.delete_item(container, children_only=True)

class ScreenFactory(ScreenHandler):
    def __init__(self, factory: Callable[[None], Screen]):
        self.factory = factory

    def show(self, container: Tag) -> None:
        screen:Screen = self.factory()
        with dpg.group(parent=container):
            screen.add_to_context()

    def hide(self, container: Tag) -> None:
        dpg.delete_item(container, children_only=True)

class ScreenItemTag(ScreenHandler):
    def __init__(self, tag: Tag):
        self.tag = tag

    def show(self, container: Tag) -> None:
        dpg.show_item(self.tag)

    def hide(self, container: Tag) -> None:
        dpg.hide_item(self.tag)


import dearpygui.dearpygui as dpg
import dearpygui_router as router

dpg.create_context()
dpg.create_viewport()
dpg.setup_dearpygui()

with dpg.window(tag="_main_window", label="Router Demo"):
    with dpg.group(horizontal=True):
        dpg.add_button(label="Go To 1", callback=lambda: router.navigate("screen1"))
        dpg.add_button(label="Go To 2", callback=lambda: router.navigate("screen2"))

    with router.router(tag="main"):
        with router.screen(route_name="screen1", is_initial=True):
            dpg.add_text("This is screen 1")

        with router.screen(route_name="screen2"):
            dpg.add_text("This is screen 2")

dpg.show_viewport()
dpg.set_primary_window("_main_window", True)
dpg.start_dearpygui()
dpg.destroy_context()
